#include "command.h"
#include "compiler.h"

using namespace Frost;

Command::Command( const s_str& sShortName, const s_str& sLongName, const s_str& sDescription ) :
    sShortName_(sShortName), sLongName_(sLongName), sDescription_(sDescription)
{
}

Command::~Command()
{
}

const s_str& Command::GetShortName() const
{
    return sShortName_;
}

const s_str& Command::GetLongName() const
{
    return sLongName_;
}

const s_str& Command::GetDescription() const
{
    return sDescription_;
}

void Command::Disable()
{
    bDisabled_ = true;
}

const s_bool& Command::IsDisabled() const
{
    return bDisabled_;
}

const s_bool& Command::IsProvided() const
{
    return bProvided_;
}

void Command::SetCompiler(s_ptr<Compiler> pCompiler)
{
    pCompiler_ = pCompiler;
}

// -----------------------------------

CommandHelp::CommandHelp() : Command("h", "help", "displays this text")
{
}

void CommandHelp::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
        return;

    bProvided_ = true;
}

void CommandHelp::Do()
{
    pCompiler_->PrintHelp();
}

// -----------------------------------

CommandNoCpp::CommandNoCpp() : Command("m", "mpp only", "tells the compiler not to call the C++ compiler")
{
}

void CommandNoCpp::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
        return;

    bProvided_ = true;
}

void CommandNoCpp::Do()
{
    pCompiler_->DoNotCompileCpp();
}

// -----------------------------------

CommandInput::CommandInput() : Command("c", "compile", "specifies the .mpp file to compile")
{
}

void CommandInput::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End() && !iter->StartsWith("-"))
    {
        sInput_ = *iter;
        bProvided_ = true;
    }
    else
    {
        throw CommandException(this, "-c must be followed by a file name", true);
    }
}

void CommandInput::Do()
{
    pCompiler_->SetCompiledFile(sInput_);
}

// -----------------------------------

CommandTarget::CommandTarget() : Command("t", "target", "specifies the name of the compilation target")
{
}

void CommandTarget::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End() && !iter->StartsWith("-"))
    {
        bProvided_ = true;
        sTarget_ = *iter;
    }
    else
    {
        throw CommandException(this, "-t must be followed by a target name", true);
    }
}

void CommandTarget::Do()
{
    pCompiler_->SetTarget(sTarget_);
}

// -----------------------------------

CommandVerbose::CommandVerbose() : Command("V", "verbose", "enables verbose mode, in which the compiler will output information about the compiling process")
{
}

void CommandVerbose::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
        return;

    bProvided_ = true;
}

void CommandVerbose::Do()
{
    pCompiler_->EnableVerboseMode();
}

// -----------------------------------

CommandNoSafeWarning::CommandNoSafeWarning() : Command("w", "no safe warning", "disable warnings that are considered 'safe' (or simple information)")
{
}

void CommandNoSafeWarning::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
        return;

    bProvided_ = true;
}

void CommandNoSafeWarning::Do()
{
    pCompiler_->DisableSafeWarnings();
}

// -----------------------------------

CommandOutput::CommandOutput() : Command("o", "output", "specifies the name of the output executable")
{
}

void CommandOutput::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End() && !iter->StartsWith("-"))
    {
        bProvided_ = true;
        sOutput_ = *iter;
    }
    else
    {
        throw CommandException(this, "-o must be followed by a file name", true);
    }
}

void CommandOutput::Do()
{
    pCompiler_->SetOutputFile(sOutput_);
}
// -----------------------------------

CommandOutputDirectory::CommandOutputDirectory() : Command("O", "output directory", "specifies the directory in which to place intermediate files")
{
}

void CommandOutputDirectory::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End() && !iter->StartsWith("-"))
    {
        bProvided_ = true;
        sOutputDirectory_ = *iter;
    }
    else
    {
        throw CommandException(this, "-O must be followed by a directory", true);
    }
}

void CommandOutputDirectory::Do()
{
    pCompiler_->SetOutputDirectory(sOutputDirectory_);
}

// -----------------------------------

CommandCompiler::CommandCompiler() : Command("C", "compiler", "specifies the name of the C++ compiler to use")
{
}

void CommandCompiler::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End() && !iter->StartsWith("-"))
    {
        bProvided_ = true;
        sCompiler_ = *iter;
    }
    else
    {
        throw CommandException(this, "-C must be followed by a compiler name", true);
    }
}

void CommandCompiler::Do()
{
    pCompiler_->SetCppCompiler(sCompiler_);
}

// -----------------------------------

CommandCompilerCommands::CommandCompilerCommands() : Command("CC", "compiler commands", "adds commands to the c++ compiler")
{
}

void CommandCompilerCommands::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End())
    {
        bProvided_ = true;
        sCompilerCommands_ = *iter;
    }
    else
    {
        throw CommandException(this, "-CC must be followed by a list of commands", true);
    }
}

void CommandCompilerCommands::Do()
{
    pCompiler_->SetCppCompilerCommands(sCompilerCommands_);
}

// -----------------------------------

CommandLinkerCommands::CommandLinkerCommands() : Command("LC", "linker commands", "adds commands to the c++ linker")
{
}

void CommandLinkerCommands::Read(s_ctnr<s_str>::const_iterator& iter, const s_ctnr<s_str>& lArgs)
{
    if (bDisabled_)
    {
        ++iter;
        return;
    }

    ++iter;
    if (iter != lArgs.End())
    {
        bProvided_ = true;
        sLinkerCommands_ = *iter;
    }
    else
    {
        throw CommandException(this, "-LC must be followed by a list of commands", true);
    }
}

void CommandLinkerCommands::Do()
{
    pCompiler_->SetCppLinkerCommands(sLinkerCommands_);
}

// -----------------------------------

CommandException::CommandException(s_ptr<Command> pCommand, const s_str& sError, const s_bool& bFatal) :
    bFatal_(bFatal)
{
    if (pCommand)
    {
        sMessage_ = "Error issuing command : " + pCommand->GetShortName() + " (" + pCommand->GetLongName() + ") :\n" +
                    "    " + sError;
    }
    else
    {
        sMessage_ = "Error : \n    " + sError;
    }
}

const s_bool& CommandException::IsFatal() const
{
    return bFatal_;
}


