#include "compiler.h"
#include "command.h"
#include "module.h"
#include "function.h"
#include "variable.h"
#include "typedef.h"
#include "enum.h"
#include "inputsource.h"

#include <frost_utils_file.h>
#include <frost_utils_directory.h>

using namespace Frost;

Compiler::Compiler() : bReady_(true),
lRestrictedKeywordList_({
    "(", ")", "::", "&", "*", "<", ">",
    "const", "mutable", "volatile", "extern", "virtual", "explicit",
    "true", "false", "this", "void", "throw", "catch",
    "enum", "class", "struct", "union", "template", "typename", "typedef",
    "for", "if", "else", "while", "do", "break", "continue", "return", "switch", "case", "default",
    "using", "namespace", "export", "float", "double", "char", "short", "int", "unsigned", "signed",
    "operator", "public", "protected", "private", "new", "delete", "inherit"
}),
lPODTypeList_({"void", "signed", "unsigned", "char", "short", "int", "long", "float", "double", "wchar_t", "size_t"}),
lTypeQualifierList_({"*", "&", "const", "mutable", "volatile"})
{
}

Compiler::~Compiler()
{
}

void Compiler::ReadCommands(const s_ctnr<s_str>& lArgs)
{
    s_bool bProject;
    s_map<s_str, s_wptr<Command>> lProvidedCommands;

    if (lArgs.GetSize() > 1)
    {
        s_ctnr<s_str>::const_iterator iterArgs;
        foreach (iterArgs, lArgs)
        {
            if (iterArgs == lArgs.Begin())
                ++iterArgs;

            const s_str& sArg = *iterArgs;
            if (sArg.StartsWith("-"))
            {
                s_str sCommand = sArg.ExtractRange(1, s_uint::INF);
                s_map<s_str, s_refptr<Command>>::iterator iter = lCommandList_.Get(sCommand);
                if (iter != lCommandList_.End())
                {
                    try
                    {
                        iter->second->Read(iterArgs, lArgs);
                        lProvidedCommands[iter->first] = iter->second;
                    }
                    catch (CommandException& e)
                    {
                        Log(e.GetDescription());

                        if (e.IsFatal())
                        {
                            bReady_ = false;
                            return;
                        }
                    }
                }
                else
                    ThrowWarning("unknown command '" + sCommand + "'");
            }
            else if (!bProject)
            {
                bProject = true;
                sCompiledProject_ = sArg;
            }
            else
                ThrowWarning("unknown command '" + sArg + "'");
        }
    }

    s_wptr<Command> pHelp = GetCommand("h");
    if (pHelp->IsProvided())
    {
        bReady_ = false;
        pHelp->Do();
        return;
    }

    s_wptr<Command> pVerbose = GetCommand("V");
    if (pVerbose->IsProvided())
    {
        pVerbose->Do();
        lProvidedCommands.Erase("V");
    }

    if (!bProject)
    {
        s_wptr<Command> pCmd = GetCommand("c");
        if (pCmd->IsProvided())
        {
            pCmd->Do();
            lProvidedCommands.Erase("c");
        }
        else
        {
            pCmd = GetCommand("p");
            if (pCmd->IsProvided())
            {
                pCmd->Do();
                lProvidedCommands.Erase("p");
            }
            else
                ThrowError("no input file, please use mcpp <project>, mcpp -c <mpp file>,\nmcpp -p <cpp file> -m <module name> or mcpp -h for help");
        }
    }

    s_map<s_str, s_wptr<Command>>::iterator iter;
    foreach (iter, lProvidedCommands)
    {
        iter->second->Do();
    }
}

void Compiler::RegisterCommand(s_refptr<Command> pCommand)
{
    pCommand->SetCompiler(this);
    lCommandList_[pCommand->GetShortName()] = pCommand;
}

s_wptr<Command> Compiler::GetCommand(const s_str& sShortName)
{
    return lCommandList_[sShortName];
}

void Compiler::PrintHelp() const
{
    s_map<s_str, s_refptr<Command>>::const_iterator iter;
    foreach (iter, lCommandList_)
    {
        s_refptr<Command> pCommand = iter->second;
        Log("  " + pCommand->GetShortName() + " : " + pCommand->GetLongName() + " : " + pCommand->GetDescription());
    }
}

void Compiler::SetCompiledFile(const s_str& sCompiled)
{
    sCompiledFile_ = sCompiled;
}

void Compiler::EnableVerboseMode()
{
    bVerbose_ = true;
}

const s_bool& Compiler::IsVerbose()
{
    return bVerbose_;
}

void Compiler::DisableSafeWarnings()
{
    bNoSafeWarning_ = true;
}

void Compiler::SetOutputFile(const s_str& sOutput)
{
    sOutputFile_ = sOutput;
}

void Compiler::SetOutputDirectory(const s_str& sOutputDirectory)
{
    sOutputDirectory_ = sOutputDirectory;
    if (!sOutputDirectory_.IsEmpty())
    {
        sOutputDirectory_.Replace("\\", "/");
        if (sOutputDirectory_.Back() != '/')
            sOutputDirectory_.PushBack('/');
    }
}

const s_str& Compiler::GetOutputDirectory() const
{
    return sOutputDirectory_;
}

void Compiler::DoNotCompileCpp()
{
    bDoNotCompileCpp_ = true;
}

void Compiler::SetCppCompiler(const s_str& sCompiler)
{
    if (sCompiler == "gcc" || sCompiler == "g++")
        sCompiler_ = "g++";
    else
        ThrowWarning("unknown compiler '" + sCompiler + "'");
}

void Compiler::SetCppCompilerCommands(const s_str& sCompilerCommands)
{
    sCompilerCommands_ = sCompilerCommands;
}

void Compiler::SetCppLinkerCommands(const s_str& sLinkerCommands)
{
    sLinkerCommands_ = sLinkerCommands;
}

void Compiler::SetTarget(const s_str& sTarget)
{
    sTargetName_ = sTarget;
}

void Compiler::ThrowError(const s_str& sError) const
{
    if (mState_.sCurrentFile_.IsEmpty())
        throw Exception("    error: " + sError);
    else
        throw Exception(mState_.sCurrentFile_ + ":" + mState_.uiCurrentLineNbr_ + "\n    error: " + sError);
}

void Compiler::ThrowWarning(const s_str& sWarning, const s_bool& bIsSafe) const
{
    if (bIsSafe && bNoSafeWarning_)
        return;

    if (mState_.sCurrentFile_.IsEmpty())
        Log("    warning: " + sWarning);
    else
        Log(mState_.sCurrentFile_ + ":" + mState_.uiCurrentLineNbr_ + "\n    warning: " + sWarning);
}

void Compiler::Compile()
{
    if (sCompiledProject_.IsEmpty())
        CompileModule(sCompiledFile_);
    else
        CompileProject(sCompiledProject_);
}

void Compiler::CompileProject(const s_str& sFile)
{
    if (!bReady_)
        return;

    if (bVerbose_) Log("Compiling project "+sFile+" ...");

    File mInput(sFile, File::I);

    if (!mInput.IsValid())
        ThrowError("cannot find file : "+sFile);

    mState_.sCurrentFile_ = sFile;
    mState_.uiCurrentLineNbr_ = 0;
    s_uint uiLineNbr;

    s_ptr<Target> pTarget;
    s_bool        bFiles;
    s_ctnr<s_wptr<Module>> lModuleList;

    while (mInput.IsValid())
    {
        ++mState_.uiCurrentLineNbr_;

        s_str sLine = mInput.GetLine();
        if (sLine.IsEmpty(true))
            continue;

        sLine.Trim(' ');
        if (sLine[0] == '#')
        {
            if (bFiles)
            {
                ThrowWarning("project command found after 'files', ignoring");
                continue;
            }

            sLine.EraseFromStart(1);
            s_ctnr<s_str> lWords = sLine.Cut(":", 1);
            s_str sName = lWords[0];
            sName.Trim(' ');
            if (lWords.GetSize() == 2)
            {
                if (sName[0] == '#')
                {
                    sName.EraseFromStart(1);
                    s_str sValue = lWords[1];
                    sValue.Trim(' ');

                    if (sName == "compiler")
                        pTarget->sCompiler_ = sValue;
                    else if (sName == "compiler_commands")
                        pTarget->sCompilerCommands_ = sValue;
                    else if (sName == "linker_commands")
                        pTarget->sLinkerCommands_ = sValue;
                    else if (sName == "output_file")
                        pTarget->sOutputFile_ = sValue;
                    else if (sName == "output_directory")
                        pTarget->sOutputDirectory_ = sValue;
                    else
                        ThrowWarning("unknown project option '"+sName+"'");
                }
                else
                {
                    pTarget = nullptr;
                    s_str sValue = lWords[1];
                    sValue.Trim(' ');

                    if (sName == "compiler" && sCompiler_.IsEmpty())
                        sCompiler_ = sValue;
                    else if (sName == "compiler_commands" && sCompilerCommands_.IsEmpty())
                        sCompilerCommands_ = sValue;
                    else if (sName == "linker_commands" && sLinkerCommands_.IsEmpty())
                        sLinkerCommands_ = sValue;
                    else if (sName == "output_file" && sOutputFile_.IsEmpty())
                        sOutputFile_ = sValue;
                    else if (sName == "output_directory" && sOutputDirectory_.IsEmpty())
                        SetOutputDirectory(sValue);
                    else if (sName == "default_target")
                        sDefaultTargetName_ = sValue;
                    else if (sName == "target")
                    {
                        if (lTargetList_.Find(sValue))
                            pTarget = &lTargetList_[sValue];
                        else
                        {
                            Target t; t.sName_ = sValue;
                            lTargetList_[sValue] = t;
                            pTarget = &lTargetList_[sValue];
                        }
                    }
                    else if (sName == "files")
                    {
                        s_ptr<Target> pTarget;
                        if (!sTargetName_.IsEmpty())
                        {
                            if (!lTargetList_.Find(sTargetName_))
                                ThrowError("unkown target '" + sTargetName_ + "'");

                            pTarget = &lTargetList_[sTargetName_];
                        }
                        else if (!sDefaultTargetName_.IsEmpty())
                        {
                            if (!lTargetList_.Find(sDefaultTargetName_))
                                ThrowError("unkown target '" + sDefaultTargetName_ + "'");

                            pTarget = &lTargetList_[sDefaultTargetName_];
                        }

                        if (!pTarget && !lTargetList_.IsEmpty())
                            ThrowWarning("no target specified, using default parameters");

                        if (pTarget)
                        {
                            if (bVerbose_) Log("Compiling target "+pTarget->sName_+" ...");

                            if (!pTarget->sCompiler_.IsEmpty() && sCompiler_.IsEmpty())
                                sCompiler_ = pTarget->sCompiler_;
                            if (!pTarget->sCompilerCommands_.IsEmpty() && sCompilerCommands_.IsEmpty())
                                sCompilerCommands_ = pTarget->sCompilerCommands_;
                            if (!pTarget->sLinkerCommands_.IsEmpty() && sLinkerCommands_.IsEmpty())
                                sLinkerCommands_ = pTarget->sLinkerCommands_;
                            if (!pTarget->sOutputFile_.IsEmpty() && sOutputFile_.IsEmpty())
                                sOutputFile_ = pTarget->sOutputFile_;
                            if (!pTarget->sOutputDirectory_.IsEmpty() && sOutputDirectory_.IsEmpty())
                                SetOutputDirectory(pTarget->sOutputDirectory_);
                        }

                        bFiles = true;
                    }
                    else
                        ThrowWarning("unknown project option '"+sName+"'");
                }
            }
        }
        else
        {
            if (!bFiles)
                ThrowError("missing 'files' command in project file");

            if (!lCompiledFileList_.Find(sLine))
            {
                uiLineNbr = mState_.uiCurrentLineNbr_;

                lModuleList.PushBack(CompileModule(sLine));

                mState_.sCurrentFile_ = sFile;
                mState_.uiCurrentLineNbr_ = uiLineNbr;
            }
            else
            {
                s_wptr<Module> pModule = lCompiledFileList_[sLine];
                if (pModule->IsPreloaded())
                {
                    mState_.sCurrentFile_ = sLine;
                    mState_.pCurrentModule_ = pModule;

                    pModule->Compile();

                    mState_.sCurrentFile_ = sFile;
                }
            }
        }
    }

    if (bDoNotCompileCpp_)
        return;

    s_ctnr<s_str> lObjectList;

    if (!lModuleList.IsEmpty())
    {
        CheckOutputDirectory();

        s_ctnr<s_wptr<Module>>::iterator iter;
        foreach (iter, lModuleList)
        {
            s_str sCppFile = GetOutputDirectory() + (*iter)->GetCreatedSourceFile();
            s_str sOutput = sCppFile;
            sOutput.EraseFromEnd(sCppFile.Cut(".").Back().GetSize());
            sOutput += "o";

            CompileCppFile(sCppFile, sOutput);
            lObjectList.PushBack(sOutput);
        }
    }
    else
        ThrowError("no module to compile");

    if (sOutputFile_.IsEmpty())
    {
        sOutputFile_ = sFile;
#ifdef WIN32
        sOutputFile_.EraseFromEnd(sFile.Cut(".").Back().GetSize());
        sOutputFile_ += "exe";
#else
#ifdef LINUX
        sOutputFile_.EraseFromEnd(sFile.Cut(".").Back().GetSize() + 1);
#else
    #error Unsupported system !
#endif
#endif
    }

    if (bVerbose_) Log("Linking project "+sFile+" to " + sOutputFile_ + "...");
    LinkFiles(lObjectList, sOutputFile_);
}

void Compiler::CompileCppFile(const s_str& sFile, const s_str& sOutput)
{
#ifdef WIN32
    if (sCompiler_.IsEmpty() || sCompiler_ == "g++")
    {
        s_str sCommand = "g++ -c " + sFile + " -o " + sOutput;
        if (!sCompilerCommands_.IsEmpty())
            sCommand += " " + sCompilerCommands_;

        if (bVerbose_) Log("  " + sCommand);

        if (system(sCommand.c_str()) != 0)
            ThrowError("unable to compile file "+sFile);
    }
#else
#ifdef LINUX
    if (sCompiler_.IsEmpty() || sCompiler_ == "g++")
    {
        s_str sCommand = "g++ -c " + sFile + " -o " + sOutput;
        if (!sCompilerCommands_.IsEmpty())
            sCommand += " " + sCompilerCommands_;

        if (bVerbose_) Log("  " + sCommand);

        if (system(sCommand.c_str()) != 0)
            ThrowError("unable to compile file "+sFile);
    }
#else
    #error Unsupported system !
#endif
#endif
}

void Compiler::LinkFiles(const s_ctnr<s_str>& lFiles, const s_str& sOutput)
{
#ifdef WIN32
    if (sCompiler_.IsEmpty() || sCompiler_ == "g++")
    {
        s_str sCommand = "g++ -o " + sOutput;

        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, lFiles)
        {
            sCommand += " " + *iter;
        }

        if (!sLinkerCommands_.IsEmpty())
            sCommand += " " + sLinkerCommands_;

        if (bVerbose_) Log("  " + sCommand);

        if (system(sCommand.c_str()) != 0)
            ThrowError("unable to link project");
    }
#else
#ifdef LINUX
    if (sCompiler_.IsEmpty() || sCompiler_ == "g++")
    {
        s_str sCommand = "g++ -o " + sOutput;
        if (!sLinkerCommands_.IsEmpty())
            sCommand += " " + sLinkerCommands_;

        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, lFiles)
        {
            sCommand += " " + *iter;
        }

        if (bVerbose_) Log("  " + sCommand);

        if (system(sCommand.c_str()) != 0)
            ThrowError("unable to link project");
    }
#else
    #error Unsupported system !
#endif
#endif
}

s_wptr<Module> Compiler::CompileModule(const s_str& sFile)
{
    if (!bReady_)
        return nullptr;

    if (bVerbose_) Log("Compiling module "+sFile+" ...");

    File mInput(sFile, File::I);

    if (!mInput.IsValid())
    {
        ThrowError("cannot find file : "+sFile);
    }

    mState_.sCurrentFile_ = sFile;
    mState_.uiCurrentLineNbr_ = 1;
    mState_.bNoTypeCheck_ = false;

    s_refptr<Module> pModule(new Module(this, sFile));
    mState_.pCurrentModule_ = pModule;
    mState_.pCurrentModule_->SetSelf(mState_.pCurrentModule_);

    ReadContent_(InputSource(mInput));

    if (bVerbose_) Log("Done.");

    if (pModule->GetName().IsEmpty())
        ThrowError("missing \'module\' instruction");

    if (bVerbose_) Log("Serializing ...");
    CheckOutputDirectory();
    pModule->Serialize();
    if (bVerbose_) Log("Done.");

    mState_.pCurrentModule_ = nullptr;

    lCompiledFileList_[sFile] = pModule;

    return pModule;
}

s_wptr<Module> Compiler::LoadModule(const s_str& sFile)
{
    if (!bReady_)
        return nullptr;

    if (bVerbose_) Log("Loading module "+sFile+" ...");

    File mInput(sFile, File::I);

    if (!mInput.IsValid())
    {
        ThrowError("cannot find file : "+sFile);
    }

    mState_.sCurrentFile_ = sFile;
    mState_.uiCurrentLineNbr_ = 1;
    mState_.bNoTypeCheck_ = true;

    s_refptr<Module> pModule(new Module(this, sFile));
    mState_.pCurrentModule_ = pModule;
    mState_.pCurrentModule_->SetSelf(mState_.pCurrentModule_);
    mState_.pCurrentModule_->SetPreloaded();

    ReadContent_(InputSource(mInput));

    if (bVerbose_) Log("Done.");

    if (pModule->GetName().IsEmpty())
        ThrowError("missing \'module\' instruction");

    mState_.pCurrentModule_ = nullptr;

    lCompiledFileList_[sFile] = pModule;

    return pModule;
}

void Compiler::ReadContent_(InputSource mInput)
{
    s_uint uiLine;

    if (mState_.bNestedClass_)
        uiLine = mState_.uiCurrentLineNbr_;
    else
        uiLine = 1;

    s_str sInstruction;
    s_ctnr<s_str> lWords;

    s_str sLine;

    s_uint uiBraceCount;
    s_uint uiSlashCount;
    s_uint uiColonCount;

    /*s_str s;

    Log("\n####\n");

    char c;
    while (mInput.IsValid())
    {
        c = mInput.Get();
        std::cout << (uint)(uchar)c << " ";

        s.PushBack(c);
    }

    std::cout << std::endl << (int)c << std::endl;

    mInput = s;*/

    // Ignore carriage returns
    mInput.Ignore(13);

    while (mInput.IsValid())
    {
        char c = mInput.Get();

        if (c == '\0' || c == 8)
            break;

        // TODO : # fix junk at eof
        if (c == -65)
            break;

        if (c == '/' && uiBraceCount == 0)
        {
            ++uiSlashCount;
            if (uiSlashCount == 2)
            {
                sInstruction.PopBack();

                if (mInput.IsValid())
                    c = mInput.Get();
                else
                    break;

                if (c == '/')
                {
                    s_str sDoc;
                    while (mInput.IsValid())
                    {
                        c = mInput.Get();

                        if (c == '\n')
                        {
                            ++uiLine;
                            break;
                        }

                        sDoc.PushBack(c);
                    }

                    if (!sDoc.IsEmpty())
                        mState_.lDocElements_.PushBack("///"+sDoc);
                }
                else if (c != '\n')
                {
                    while (mInput.IsValid())
                    {
                        c = mInput.Get();

                        if (c == '\n')
                        {
                            ++uiLine;
                            break;
                        }
                    }
                }

                if (!mInput.IsValid())
                {
                    if (!sInstruction.IsEmpty())
                        lWords.PushBack(sInstruction);

                    Evaluate_(lWords);
                    mState_.uiCurrentLineNbr_ = uiLine;
                    lWords.Clear();
                    sInstruction.Clear();
                    break;
                }

                if (lWords.IsEmpty())
                    mState_.uiCurrentLineNbr_ = uiLine;

                uiSlashCount = 0;
                continue;
            }
            else
            {
                sInstruction.PushBack(c);
                continue;
            }
        }
        else if (c == '*' && uiSlashCount == 1)
        {
            sInstruction.PopBack();
            s_bool bDoc;
            s_str sDoc;

            s_bool bFoundStar;

            if (mInput.IsValid())
                c = mInput.Get();
            else
                break;

            if (c == '*')
            {
                sDoc = "/**";
                bDoc = true;
                bFoundStar = true;
            }

            s_bool bClosed;
            while (mInput.IsValid())
            {
                c = mInput.Get();

                if (bDoc)
                    sDoc.PushBack(c);

                if (c == '/')
                {
                    if (bFoundStar)
                    {
                        bClosed = true;
                        break;
                    }
                }
                else if (c == '*')
                {
                    bFoundStar = true;
                }
                else
                {
                    bFoundStar = false;

                    if (c == '\n')
                        ++uiLine;
                }
            }

            if (lWords.IsEmpty())
                mState_.uiCurrentLineNbr_ = uiLine;

            if (!mInput.IsValid())
            {
                if (!bClosed)
                    ThrowError("missing comment closing sign '*/'");
                else
                {
                    if (!sInstruction.IsEmpty())
                        lWords.PushBack(sInstruction);

                    Evaluate_(lWords);
                    mState_.uiCurrentLineNbr_ = uiLine;
                    lWords.Clear();
                    sInstruction.Clear();
                    break;
                }
            }

            if (bDoc)
                mState_.lDocElements_.PushBack(sDoc);

            uiSlashCount = 0;
            continue;
        }
        else
            uiSlashCount = 0;

        if (c == '"')
        {
            sInstruction.PushBack(c);

            s_bool bEscape;
            while (mInput.IsValid())
            {
                c = mInput.Get();

                if (c == '\n')
                    ++uiLine;

                if (!bEscape && c == '\\')
                    bEscape = true;
                else if (!bEscape && c == '"')
                {
                    sInstruction.PushBack(c);
                    break;
                }
                else
                {
                    bEscape = false;
                    sInstruction.PushBack(c);
                }
            }

            if (!mInput.IsValid())
                ThrowError("end of file reached while parsing string constant");

            continue;
        }

        if (c == ':' && uiBraceCount == 0)
        {
            if (uiColonCount == 0)
            {
                if (!sInstruction.IsEmpty())
                    lWords.PushBack(sInstruction);
                sInstruction.Clear();
            }

            ++uiColonCount;
        }
        else if (uiColonCount != 0)
        {
            lWords.PushBack(sInstruction);
            sInstruction.Clear();

            if (uiColonCount == 1)
            {
                if (lWords.GetSize() < 2)
                    ThrowError("ill formed statement : ':'");

                if (lWords[lWords.GetSize() - 2] != ")")
                {
                    Evaluate_(lWords);
                    mState_.uiCurrentLineNbr_ = uiLine;
                    lWords.Clear();
                }
            }

            uiColonCount = 0;
        }

        if (c == ';' && uiBraceCount == 0)
        {
            if (!sInstruction.IsEmpty())
                lWords.PushBack(sInstruction);

            Evaluate_(lWords);
            mState_.uiCurrentLineNbr_ = uiLine;
            lWords.Clear();
            sInstruction.Clear();
        }
        else if (c == '{')
        {
            if (uiBraceCount == 0)
            {
                if (!sInstruction.IsEmpty())
                    lWords.PushBack(sInstruction);

                lWords.PushBack("{");
                sInstruction.Clear();
            }
            else
                sInstruction.PushBack('{');

            ++uiBraceCount;
        }
        else if (c == '}')
        {
            --uiBraceCount;
            if (uiBraceCount == 0)
            {
                if (!sInstruction.IsEmpty())
                    lWords.PushBack(sInstruction);

                lWords.PushBack("}");
                Evaluate_(lWords);
                mState_.uiCurrentLineNbr_ = uiLine;
                lWords.Clear();
                sInstruction.Clear();
            }
            else
                sInstruction.PushBack('}');
        }
        else if ((c == '(' || c == ')') && uiBraceCount == 0)
        {
            if (!sInstruction.IsEmpty())
                lWords.PushBack(sInstruction);

            sInstruction.Clear();
            sInstruction.PushBack(c);
            lWords.PushBack(sInstruction);
            sInstruction.Clear();
        }
        else if ((c == ',' || c == '<' || c == '>' || c == '*' || c == '&') && uiBraceCount == 0)
        {
            if (!sInstruction.IsEmpty())
                lWords.PushBack(sInstruction);

            lWords.PushBack(c);
            sInstruction.Clear();
        }
        else if (c != ' ' && c != '\t' && c != '\n')
        {
            sInstruction.PushBack(c);
        }
        else
        {
            if (c == '\n')
            {
                ++uiLine;
                if (lWords.IsEmpty())
                    mState_.uiCurrentLineNbr_ = uiLine;
            }

            if (uiBraceCount != 0)
            {
                sInstruction.PushBack(c);
            }
            else if (!sInstruction.IsEmpty())
            {
                lWords.PushBack(sInstruction);
                sInstruction.Clear();
            }
        }
    }

    if (!sInstruction.IsEmpty(true) || !lWords.IsEmpty())
    {
        s_str sError;
        if (mState_.bNestedClass_)
            sError = "unexpected end of "+mState_.pCurrentModule_->GetCurrentClass()->GetClassType()+", missing semicolon ?";
        else
            sError = "unexpected end of file, missing semicolon ?";

        if (bVerbose_)
        {
            s_ctnr<s_uint> lInst;
            s_str::iterator iter;
            foreach (iter, sInstruction)
                lInst.PushBack((uint)(uchar)*iter);

            sError += "\n    "+lInst;
        }

        ThrowError(sError);
    }
}

void Compiler::Evaluate_(const s_ctnr<s_str>& lWords)
{
    if (lWords.IsEmpty())
        return;

    /*s_str sTotal;
    Log("-------------------------");
    s_ctnr<s_str>::const_iterator iterWords;
    foreach (iterWords, lWords)
    {
        sTotal << *iterWords << "\n";
    }
    Log(sTotal);
    Log("-------------------------");*/

    s_str sStart = lWords[0];

    // TODO : allow "namespace" ?

    if (sStart == "module")
    {
        if (lWords.GetSize() == 1)
            ThrowError("missing module name after \'module\'");

        if (mState_.bNestedClass_)
            ThrowError("'module' command forbidden within classes");

        s_ctnr<s_str> lName = lWords;

        s_bool bClass;
        if (lWords[1] == "class")
        {
            bClass = true;
            if (lWords.GetSize() == 2)
                ThrowError("missing module name after \'module class\'");

            lName.PopFront(2);
            mState_.pCurrentModule_->CreateMainClass();
        }
        else
            lName.PopFront(1);

        TypeName mName = ReadModuleName_(lName);
        mState_.pCurrentModule_->SetName(mName);
        mState_.pCurrentModule_->SetDocElements(mState_.lDocElements_);
        mState_.lDocElements_.Clear();

        lModuleList_[mState_.pCurrentModule_->GetFullName()] = mState_.pCurrentModule_.Lock();

        mState_.lContextStack_ = mName.lNamespaces;
        if (bClass)
            mState_.lContextStack_.PushBack(mName.sType);
    }
    else if (mState_.pCurrentModule_->GetName().IsEmpty())
    {
        ThrowError("cannot use '"+sStart+"' before \'module\' instruction");
    }
    else if (sStart == "set")
    {
        if (lWords.GetSize() == 1)
            ThrowError("expected instruction after \'set\'");

        s_str sInst = lWords[1];

        if (sInst == "header")
        {
            if (lWords.GetSize() == 2)
                ThrowError("expected header name after \'set header\'");

            s_str sHeader = lWords[2];

            if (sHeader == "<")
            {
                if (lWords.GetSize() < 5 || lWords[4] != ">")
                    ThrowError("ill formed header name");

                sHeader += lWords[3] + lWords[4];

                if (lWords.GetSize() > 5)
                    ThrowError("unexpected content after \'set header "+sHeader+"\'");
            }
            else if (lWords.GetSize() > 3)
                ThrowError("unexpected content after \'set header "+sHeader+"\'");

            mState_.pCurrentModule_->SetHeaderFile(sHeader);
        }
    }
    else if (sStart == "using")
    {
        if (lWords.GetSize() == 1)
            ThrowError("missing content after 'using'");

        if (mState_.bNestedClass_)
            ThrowError("'using' command forbidden within classes");

        if (lWords[1] == "{")
        {
            if (lWords[2] != "}")
                mState_.pCurrentModule_->UsesContent(lWords[2]);
        }
        else if (lWords[1] == "namespace")
        {
            if (lWords.GetSize() == 2)
                ThrowError("missing content after 'using namespace'");

            s_ctnr<s_str> lNamespaces = lWords;
            lNamespaces.PopFront(2);
            mState_.pCurrentModule_->UsesNamespace(lNamespaces);
        }
        else if (lWords[1] == "module")
        {
            s_ctnr<s_str> lModule = lWords;
            lModule.PopFront(2);
            TypeName mName = ReadModuleName_(lModule);

            s_wptr<Module> pModule;
            if (!lModuleList_.Find(mName.sFullType))
                pModule = LoadModule_(mName);
            else
                pModule = lModuleList_[mName.sFullType];

            mState_.pCurrentModule_->UsesModule(pModule);
        }
        else
        {
            s_ctnr<s_str> lType = lWords;
            lType.PopFront();
            mState_.pCurrentModule_->UsesType(lType);
        }
    }
    else if (sStart == "export")
    {
        if (lWords.GetSize() == 1)
            ThrowError("missing content after \'export\'");

        if (mState_.bNestedClass_)
            ThrowError("'export' command forbidden within nested classes");

        if (lWords[1] == "{")
        {
            if (lWords[2] != "}")
                mState_.pCurrentModule_->Exports(lWords[2]);
        }
        else
            ThrowError("expected brace enclosed content avec 'export'");
    }
    else if (sStart == "public" || sStart == "protected" || sStart == "private")
    {
        if (lWords.GetSize() == 1 || lWords[1] != ":")
            ThrowError("expected ':' after '"+sStart+"'");

        if (mState_.bUnion_)
            ThrowError("there is no scope in unions");

        mState_.pCurrentModule_->SetScope(sStart);
    }
    else if (sStart == "inherit")
    {
        if (lWords.GetSize() == 1)
            ThrowError("expected class name after 'inherit'");

        if (mState_.bUnion_)
            ThrowError("unions cannot inherit from other classes");

        s_ctnr<s_str> lTemp = lWords;
        lTemp.PopFront();
        mState_.pCurrentModule_->AddInheritance(ReadType_(lTemp));
    }
    else if (sStart == "class" || sStart == "struct" || sStart == "union")
    {
        ReadClass_(lWords);
    }
    else if (sStart == "typedef" || (lWords.GetSize() > 1 && lWords[1] == "typedef"))
    {
        if (lWords.GetSize() < 3 || (lWords[1] == "typedef" && lWords.GetSize() < 4))
            ThrowError("ill formed 'typedef' statement");

        if (mState_.bUnion_)
            ThrowError("unions cannot contain typedefs");

        mState_.pCurrentModule_->AddTypedef(ReadTypedef_(lWords));
    }
    else if (sStart == "enum" || (lWords.GetSize() > 1 && lWords[1] == "enum"))
    {
        if (sStart == "enum")
        {
            if (lWords.GetSize() == 1)
                ThrowError("expected enum name");

            if (lWords.GetSize() < 5)
                ThrowError("expected enum body after enum name");
        }
        else
        {
            if (lWords.GetSize() == 2)
                ThrowError("expected enum name");

            if (lWords.GetSize() < 6)
                ThrowError("expected enum body after enum name");
        }

        if (mState_.bUnion_)
            ThrowError("unions cannot contain enums");

        mState_.pCurrentModule_->AddEnum(ReadEnum_(lWords));
    }
    else if (sStart == "template")
    {
        if (lWords.GetSize() < 3 || lWords[1] != "<")
            ThrowError("ill formed template parameters, missing braces '<' '>'");
        if (lWords.GetSize() < 4)
            ThrowError("missing content after template parameters");

        s_ctnr<s_str> lTemplateWords;

        s_uint uiOpen;
        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, lWords)
        {
            lTemplateWords.PushBack(*iter);

            if (*iter == "<")
                ++uiOpen;
            else if (*iter == ">")
            {
                --uiOpen;
                if (uiOpen == 0)
                    break;
            }
        }

        mState_.mTemplateParams_ = ReadTemplateParams_(lTemplateWords);
        mState_.lTemplateTypeStack_.PushBack(mState_.mTemplateParams_.lTemplateTypes);

        s_ctnr<s_str> lRemaining = lWords;
        lRemaining.PopFront(lTemplateWords.GetSize());

        if (lRemaining[0] == "module")
        {
            if (lRemaining.GetSize() == 1)
                ThrowError("missing module name after \'module\'");

            if (mState_.bNestedClass_)
                ThrowError("'module' command forbidden within classes");

            s_ctnr<s_str> lName = lRemaining;

            if (lRemaining[1] != "class")
                ThrowError("template modules can only be class/struct modules");

            if (lRemaining.GetSize() == 2)
                ThrowError("missing module name after \'module class\'");

            lName.PopFront(2);
            mState_.pCurrentModule_->CreateMainClass();
            mState_.pCurrentModule_->GetCurrentClass()->SetTemplateParams(mState_.mTemplateParams_);
            mState_.mTemplateParams_ = TemplateParams();

            mState_.pCurrentModule_->SetName(ReadModuleName_(lName));
            mState_.pCurrentModule_->SetDocElements(mState_.lDocElements_);
            mState_.lDocElements_.Clear();

            lModuleList_[mState_.pCurrentModule_->GetFullName()] = mState_.pCurrentModule_.Lock();
        }
        else if (lRemaining[0] == "class" || lRemaining[0] == "struct")
        {
            ReadClass_(lRemaining, true);
            mState_.lTemplateTypeStack_.PopBack();
        }
        else if (lRemaining[0] == "union" || lRemaining[0] == "typedef" || lRemaining[0] == "enum")
        {
            ThrowError(lRemaining[0]+"s cannot have template parameters");
        }
        else
        {
            s_bool bOpen =  lRemaining.Find("{");
            s_bool bClose = lRemaining.Find("}");

            if (bOpen || bClose || lRemaining[0] == "virtual")
            {
                if ((bOpen && !bClose) || (!bOpen && bClose))
                    ThrowError("ill formed function, missing brace");

                if (mState_.bUnion_)
                    ThrowError("unions cannot contain member functions");

                mState_.pCurrentModule_->AddFunction(ReadFunction_(lRemaining, true));
                mState_.lTemplateTypeStack_.PopBack();
            }
            else
            {
                ThrowError("template variables are not allowed");
            }
        }
    }
    else
    {
        s_bool bOpen =  lWords.Find("{");
        s_bool bClose = lWords.Find("}");

        if (bOpen || bClose || lWords[0] == "virtual")
        {
            if ((bOpen && !bClose) || (!bOpen && bClose))
                ThrowError("ill formed function, missing brace");

            if (mState_.bUnion_)
                ThrowError("unions cannot contain member functions");

            mState_.pCurrentModule_->AddFunction(ReadFunction_(lWords));
        }
        else
        {
            mState_.pCurrentModule_->AddVariable(ReadVariable_(lWords));
        }
    }
}

void Compiler::ReadClass_(const s_ctnr<s_str>& lWords, const s_bool& bIsTemplate)
{
    if (lWords.GetSize() == 1)
        ThrowError("expected "+lWords[0]+" name");

    if (lWords.GetSize() < 5 && lWords.GetSize() != 2)
        ThrowError("expected "+lWords[0]+" body after "+lWords[0]+" name");

    if (mState_.bUnion_)
        ThrowError("unions cannot contain nested "+lWords[0]);

    // TODO : check class name (lWords[1])

    s_bool bWasNestedClass = mState_.bNestedClass_;

    if (!bWasNestedClass)
        mState_.bNestedClass_ = true;

    s_wptr<Class> pClass = mState_.pCurrentModule_->BeginClass(lWords[1], lWords[0]);

    if (bIsTemplate)
    {
        pClass->SetTemplateParams(mState_.mTemplateParams_);
        mState_.mTemplateParams_ = TemplateParams();
    }

    if (lWords[0] == "union")
        mState_.bUnion_ = true;

    mState_.pCurrentModule_->SetDocElements(mState_.lDocElements_);
    mState_.lDocElements_.Clear();

    if (lWords.GetSize() == 5)
    {
        mState_.lContextStack_.PushBack(lWords[1]);
        ReadContent_(InputSource(lWords[3]));
        mState_.lContextStack_.PopBack();
    }

    mState_.pCurrentModule_->EndClass();
    mState_.bUnion_ = false;

    if (!bWasNestedClass)
        mState_.bNestedClass_ = false;
}

s_bool Compiler::FindModuleFile(s_str& sFileName)
{
    if (File::Exists(sFileName))
        return true;

    if (File::Exists("include/"+sFileName))
    {
        sFileName = "include/"+sFileName;
        return true;
    }

    return false;
}

s_wptr<Module> Compiler::LoadModule_(const TypeName& mName)
{
    State mOldState = mState_;

    s_str sFile = Module::MakeFileName(mName);
    if (!FindModuleFile(sFile))
        ThrowError("couldn't find module '"+mName.sFullType+"', expected file name : '"+sFile+"'");

    s_wptr<Module> pModule = LoadModule(sFile);

    mState_ = mOldState;

    return pModule;
}

TypeName Compiler::ReadModuleName_(const s_ctnr<s_str>& lWords) const
{
    TypeName mName;

    if (lWords.Find("::"))
    {
        s_ctnr<s_str>::const_iterator iter;
        foreach (iter, lWords)
        {
            mName.sFullType += *iter;

            if (*iter != "::")
            {
                s_str sTemp = *iter;
                ++iter;
                if (iter == lWords.End())
                {
                    mName.sType = sTemp;
                    mName.sNakedType = sTemp;
                    mName.sNestedType = sTemp;
                    break;
                }
                else
                {
                    if (*iter != "::")
                        ThrowError("ill formed module name, synthax is : module [class] nsp::name [: \"header file\"]");

                    mName.lNamespaces.PushBack(sTemp);

                    if (mName.sNamespace.IsEmpty())
                        mName.sNamespace = sTemp;
                    else
                        mName.sNamespace += "::" + sTemp;

                    mName.sFullType += "::";
                }
            }
        }
    }
    else
    {
        mName.sFullType = mName.sType = lWords[0];

        if (lWords.GetSize() > 2)
            ThrowWarning("ignored input after module name");
    }

    mName.sFullNakedType = mName.sFullType;
    mName.sNakedType = mName.sNestedType = mName.sType;

    return mName;
}


TemplateParams Compiler::ReadTemplateParams_(const s_ctnr<s_str>& lWords) const
{
    TemplateParams mParams;

    s_str sTemplate = "template<";
    s_uint uiOpened = 1;
    s_ctnr<s_str>::const_iterator iter = lWords.Begin() + 2;
    while (iter != lWords.End())
    {
        if (sTemplate == "template<" || *iter == ">" || *iter == "<")
            sTemplate += *iter;
        else
            sTemplate += " " + *iter;

        if (*iter == "<")
            ++uiOpened;
        if (*iter == ">")
        {
            --uiOpened;
            if (uiOpened == 0)
                break;
        }

        ++iter;
    }

    if (uiOpened != 0)
        ThrowError("ill formed template argument, missing closing '>'");

    mParams.sSerialized = sTemplate;

    // FIXME : test template<class T, class U = Test<float, float>, class V = Test<double, double>>
    sTemplate.EraseFromStart(9);
    sTemplate.EraseFromEnd(1);
    if (!sTemplate.IsEmpty())
    {
        s_ctnr<s_str> lParams = sTemplate.Cut(",");
        s_ctnr<s_str>::iterator iterParam;
        foreach (iterParam, lParams)
        {
            s_ctnr<s_str> lWords = iterParam->Cut(" ");
            if (lWords[0] == "typename" || lWords[0] == "class")
            {
                if (lWords.GetSize() == 1)
                    ThrowError("ill formed template argument, missing type name after '"+lWords[0]+"'");

                s_str sType = lWords[1];
                if (sType.Find("="))
                    sType = sType.Cut("=").Front();

                mParams.lTemplateTypes.PushBack(sType);
            }
        }
    }

    return mParams;
}

Function Compiler::ReadFunction_(const s_ctnr<s_str>& lWords, const s_bool& bIsTemplate) const
{
    //Log("#function : "+lWords);
    s_uint uiTypeStart = 0;

    Function mFunc;

    mFunc.lDocElements = mState_.lDocElements_;
    mState_.lDocElements_.Clear();

    if (bIsTemplate)
    {
        mFunc.mTemplateParams = mState_.mTemplateParams_;
        mState_.mTemplateParams_ = TemplateParams();

        while (lWords[uiTypeStart] == ",")
        {
            ++uiTypeStart;
            if (lWords[uiTypeStart] == "using")
            {
                s_ctnr<s_str> lModule;

                ++uiTypeStart;
                while (lWords[uiTypeStart] != ",")
                {
                    lModule.PushBack(lWords[uiTypeStart]);
                    ++uiTypeStart;
                }

                mFunc.lDependencies.PushBack(ReadModuleName_(lModule));
            }
        }
    }

    if (lWords[uiTypeStart] == "using")
    {
        if (bIsTemplate)
            ThrowError("missing comma after template parameters");
        else
            ThrowError("'using' command is only allowed for template functions");
    }

    if (lWords[uiTypeStart] == "virtual")
    {
        mFunc.bVirtual = true;
        ++uiTypeStart;
    }
    else if (lWords[uiTypeStart] == "static")
    {
        mFunc.bStatic = true;
        ++uiTypeStart;
    }
    else if (lWords[uiTypeStart] == "global")
    {
        // TODO : change "global" to be really global (no namespace) ?
        if (mState_.bNestedClass_)
            ThrowError("declaring a 'global' function is forbidden within classes");

        mFunc.bGlobal = true;
        ++uiTypeStart;
    }
    else if (lWords[uiTypeStart] == "new")
    {
        if (!mState_.pCurrentModule_->GetCurrentClass())
            ThrowError("cannot declare constructor at global scope");

        mFunc.sName = mState_.pCurrentModule_->GetCurrentClass()->GetType().sType;
        mFunc.bConstructor = true;
        ++uiTypeStart;
    }
    else if (lWords[uiTypeStart] == "delete")
    {
        if (!mState_.pCurrentModule_->GetCurrentClass())
            ThrowError("cannot declare destructor at global scope");

        mFunc.sName = mState_.pCurrentModule_->GetCurrentClass()->GetType().sType;
        mFunc.bDestructor = true;
        ++uiTypeStart;
    }

    s_uint uiEnd = lWords.FindPos("{");
    if (uiEnd <= uiTypeStart)
    {
        ThrowError("ill formed function, braces must be placed after the function declaration");
    }

    if (!uiEnd.IsValid())
    {
        mFunc.bPureVirtual = true;
        uiEnd = lWords.GetSize();
    }
    else
        mFunc.sContent = lWords[uiEnd + 1];

    if (lWords[uiEnd - 1] == "const")
    {
        if (!mState_.pCurrentModule_->GetCurrentClass())
            ThrowError("non member functions cannot be declared 'const'");

        mFunc.bConst = true;
        --uiEnd;
    }

    s_bool bInit = lWords.Find(":");
    if (!mFunc.bConstructor || !bInit)
    {
        if (lWords[uiEnd - 1] != ")")
            ThrowError("ill formed function, only \'const\' is allowed just after the function parameters, got : \'"+lWords[uiEnd-1]+"\'");

        --uiEnd;
    }
    else
    {
        s_ctnr<s_str>::const_iterator iter = lWords.Get(":");
        uiEnd = iter - lWords.Begin() - 1;
        ++iter;
        if (iter == lWords.End())
            ThrowError("ill formed constructor, missing content after ':'");

        while (iter != lWords.End())
        {
            if (*iter == "{")
                break;

            if (*iter == ",")
            {
                ++iter;
                continue;
            }

            Variable v;
            v.sName = *iter;
            ++iter;
            if (iter == lWords.End())
                ThrowError("ill formed constructor, missing initialization for variable '"+v.sName+"'");
            if (*iter != "(")
                ThrowError("ill formed constructor, missing opening parenthesis, got '"+(*iter)+"'");

            ++iter;
            s_uint uiParenthesis = 1;
            while (iter != lWords.End())
            {
                if (*iter == "(")
                    ++uiParenthesis;
                else if (*iter == ")")
                {
                    --uiParenthesis;
                    if (uiParenthesis == 0)
                        break;
                }
                else
                {
                    if (v.sValue.IsEmpty())
                        v.sValue = *iter;
                    else
                        v.sValue += " " + *iter;
                }

                ++iter;
            }

            if (uiParenthesis != 0)
                ThrowError("ill formed constructor, missing closing parenthesis");

            mFunc.lInit.PushBack(v);

            if (iter != lWords.End())
                ++iter;
        }
    }

    s_uint uiParamsEnd = uiEnd;

    s_uint i = uiEnd;
    s_uint uiOpened = 1;
    while (i > 0 && uiOpened != 0)
    {
        --i;
        if (lWords[i] == ")")
            ++uiOpened;
        if (lWords[i] == "(")
            --uiOpened;
    }

    if (uiOpened != 0)
        ThrowError("ill formed function, missing " + s_str((uiOpened > 0) ? "closing" : "opening") + " brace");

    s_uint uiParamsStart = i + 1;

    uiEnd = i;

    if (!mFunc.bConstructor && !mFunc.bDestructor)
    {
        mFunc.sName = lWords[uiEnd - 1];
        CheckName_(mFunc.sName);
        --uiEnd;

        s_uint uiTypeEnd = uiEnd;

        s_ctnr<s_str> lType;
        for (s_uint j = uiTypeStart; j < uiTypeEnd; ++j)
            lType.PushBack(lWords[j]);

        mFunc.mReturnType = ReadType_(lType);
    }

    s_uint uiLastID = uiParamsStart;
    for (s_uint j = uiParamsStart; j < uiParamsEnd; ++j)
    {
        if (lWords[j] == "," || j == uiParamsEnd-1)
        {
            if (j == uiParamsEnd-1)
                ++j;

            Variable v;

            s_bool bDefaultValue;
            for (s_uint k = uiLastID; k < j; ++k)
            {
                if (lWords[k] == "=")
                {
                    bDefaultValue = true;

                    v.sName = lWords[k-1];
                    CheckName_(v.sName);

                    for (s_uint l = k+1; l < j; ++l)
                        v.sValue += lWords[l];
                    s_ctnr<s_str> lType;
                    for (s_uint l = uiLastID; l < k-1; ++l)
                        lType.PushBack(lWords[l]);
                    v.mType = ReadType_(lType);

                    break;
                }
            }

            if (!bDefaultValue)
            {
                v.sName = lWords[j-1];
                CheckName_(v.sName);

                s_ctnr<s_str> lType;
                for (s_uint l = uiLastID; l < j-1; ++l)
                    lType.PushBack(lWords[l]);
                v.mType = ReadType_(lType);
            }

            uiLastID = j+1;
            mFunc.lParams.PushBack(v);
        }
    }

    /*Log("----------------------------");
    Log("type   : "+ mFunc.sType);
    Log("name   : "+ mFunc.sName);
    Log("params : "+ mFunc.lParams);*/

    return mFunc;
}

Variable Compiler::ReadVariable_(const s_ctnr<s_str>& lWords) const
{
    //Log("#variable : "+lWords);
    Variable v;

    v.lDocElements = mState_.lDocElements_;
    mState_.lDocElements_.Clear();

    s_uint uiStart = 1;
    s_uint uiSize = lWords.GetSize();

    if (lWords[0] == "global")
    {
        if (mState_.bNestedClass_)
            ThrowError("declaring a 'global' variable is forbidden within classes");

        v.bGlobal = true;
    }
    else if (lWords[0] == "static")
        v.bStatic = true;
    else
        uiStart = 0;

    s_bool bDefaultValue;
    for (s_uint k = uiStart; k < uiSize; ++k)
    {
        if (lWords[k] == "=")
        {
            bDefaultValue = true;

            v.sName = lWords[k-1];
            CheckName_(v.sName);

            for (s_uint l = k+1; l < uiSize; ++l)
                v.sValue += lWords[l];
            s_ctnr<s_str> lType;
            for (s_uint l = uiStart; l < k-1; ++l)
                lType.PushBack(lWords[l]);
            v.mType = ReadType_(lType);

            break;
        }
    }

    if (!bDefaultValue)
    {
        v.sName = lWords[uiSize-1];
        CheckName_(v.sName);

        s_ctnr<s_str> lType;
        for (s_uint l = uiStart; l < uiSize-1; ++l)
            lType.PushBack(lWords[l]);
        v.mType = ReadType_(lType);
    }

    return v;
}

Typedef Compiler::ReadTypedef_(const s_ctnr<s_str>& lWords) const
{
    //Log("\n#typedef : "+lWords);
    Typedef t;

    t.lDocElements = mState_.lDocElements_;
    mState_.lDocElements_.Clear();

    s_uint uiStart = 1;

    if (lWords[1] == "typedef")
    {
        if (lWords[0] == "global")
        {
            t.bGlobal = true;
            uiStart = 2;
        }
        else
            ThrowError("unsupported keyword before typedef : '"+lWords[0]+"'");
    }

    if (lWords.Find("(", uiStart))
    {
        t.bFunction = true;

        s_uint uiEndReturn = lWords.FindPos("(");
        s_ctnr<s_str> lType = lWords;
        lType.PopFront(uiStart);
        lType.EraseRange(uiEndReturn-1, s_uint::INF);
        t.mReturnType = ReadType_(lType);

        if (lWords.GetSize() < uiEndReturn+4)
            ThrowError("ill formed function typedef, missing typedef name");

        s_uint uiEndName = lWords.FindPos(")", uiEndReturn+1);
        if (!uiEndName.IsValid())
            ThrowError("ill formed function typedef, missing closing parenthesis");

        if (lWords[uiEndReturn+1] == "*")
        {
            if (uiEndName != uiEndReturn+3)
                ThrowError("ill formed function typedef, wrong position of the '*' ?");

            t.mType.sType = lWords[uiEndReturn+2];
            CheckName_(t.mType.sType);
        }
        else
        {
            s_uint uiStartName = uiEndReturn+1;

            s_str sTemp;
            for (s_uint i = uiStartName; i < uiEndName; ++i)
                sTemp += lWords[i];

            t.mType.sType = sTemp.Cut("::").Back();
            if (t.mType.sType[0] != '*')
                ThrowError("ill formed function typedef, missing '*' before typedef name");
            t.mType.sType.EraseFromStart(1);

            sTemp.EraseFromEnd(t.mType.sType.GetSize()+3);
            t.mMembership = ReadType_(sTemp);
        }

        s_uint uiStartParams = uiEndName+1;

        if (uiStartParams >= lWords.GetSize())
            ThrowError("ill formed function typedef, missing function arguments");

        if (lWords[uiStartParams] != "(")
            ThrowError("ill formed function typedef, missing opening parenthesis");

        ++uiStartParams;

        s_uint uiEndParams;
        s_ctnr<s_str>::const_iterator iter = lWords.End();
        while (iter != lWords.Begin())
        {
            --iter;
            if (*iter == ")")
            {
                uiEndParams = iter - lWords.Begin();
                break;
            }
        }

        if (uiEndParams == uiEndName)
            ThrowError("ill formed function typedef, missing closing parenthesis");

        lType.Clear();
        for (s_uint i = uiStartParams; i < uiEndParams; ++i)
        {
            if (lWords[i] != ",")
                lType.PushBack(lWords[i]);
            else
            {
                t.lParams.PushBack(ReadType_(lType));
                lType.Clear();
            }
        }

        if (!lType.IsEmpty())
            t.lParams.PushBack(ReadType_(lType));

        if (uiEndParams < lWords.GetSize()-2)
            ThrowError("ill formed function typedef, unexpected content after parameters");
        if (uiEndParams < lWords.GetSize()-1)
        {
            if (lWords.Back() == "const")
                t.bConst = true;
            else
                ThrowError("ill formed function typedef, unexpected content after parameters");
        }

        /*Log("name : "+t.mType);
        Log("rep  : "+t.mReturnType+" "+t.lParams+(t.bConst ? " const" : ""));*/
    }
    else
    {
        t.mType.sType = lWords.Back();
        CheckName_(t.mType.sType);

        s_ctnr<s_str> lType = lWords;
        lType.PopFront(uiStart);
        lType.PopBack();

        t.mOldType = ReadType_(lType);

        /*Log("name : "+t.mType);
        Log("rep  : "+t.mOldType);*/
    }

    return t;
}

Enum Compiler::ReadEnum_(const s_ctnr<s_str>& lWords) const
{
    //Log("#enum : "+lWords);
    Enum e;

    e.lDocElements = mState_.lDocElements_;
    mState_.lDocElements_.Clear();

    s_uint uiStart = 1;

    if (lWords[1] == "enum")
    {
        if (lWords[0] == "global")
        {
            e.bGlobal = true;
            uiStart = 2;
        }
        else
            ThrowError("unsupported keyword before enum : '"+lWords[0]+"'");
    }

    e.mType.sType = lWords[uiStart];

    ++uiStart;

    if (lWords[uiStart] != "{")
        ThrowError("ill formed enum : missing opening brace");
    if (lWords.Back() != "}")
        ThrowError("ill formed enum : missing closing brace");

    ++uiStart;

    e.sContent = lWords[uiStart];

    return e;
}

void Compiler::CheckName_(const s_str& sName) const
{
    if (lRestrictedKeywordList_.Find(sName))
        ThrowError("'"+sName+"' is a restricted keyword");

    s_str::const_iterator iter;
    foreach (iter, sName)
    {
        if (!s_str::IsAlphaNumeric(*iter) && *iter != '_')
            ThrowError("'"+sName+"' is forbidden : names can only contain alphanumeric characters");
    }
}

s_bool Compiler::IsTemplateType_(const s_str& sType) const
{
    s_ctnr<s_ctnr<s_str>>::const_iterator iter1;
    foreach (iter1, mState_.lTemplateTypeStack_)
    {
        if (iter1->Find(sType))
            return true;
    }

    return false;
}

TypeName Compiler::ReadType_(const s_ctnr<s_str>& lWords) const
{
    //Log("\n"+lWords);
    TypeName t;

    if (lWords.GetSize() == 1)
    {
        t.sType = lWords[0];

        if (!lPODTypeList_.Find(t.sType) && !IsTemplateType_(t.sType) && !mState_.bNoTypeCheck_)
        {
            mState_.pCurrentModule_->DecorateType(t, mState_.lContextStack_);
        }
        else
            t.sFullType = t.sFullNakedType = t.sNestedType = t.sNakedType = t.sType;

        return t;
    }

    s_ctnr<s_str> lModWords;
    s_str sWord;
    s_ctnr<s_str>::const_iterator iter;
    foreach (iter, lWords)
    {
        if (*iter == "::" || sWord.IsEmpty() || sWord.EndsWith("::"))
            sWord += *iter;
        else
        {
            if (!sWord.IsEmpty())
                lModWords.PushBack(sWord);
            sWord = *iter;
        }
    }

    if (!sWord.IsEmpty())
        lModWords.PushBack(sWord);

    if (lModWords.GetSize() == 1)
    {
        t.sType = lModWords[0];

        if (!lPODTypeList_.Find(t.sType) && !IsTemplateType_(t.sType) && !mState_.bNoTypeCheck_)
        {
            mState_.pCurrentModule_->DecorateType(t, mState_.lContextStack_);
        }
        else
        {
            if (IsTemplateType_(t.sType))
                t.bIsTemplateParam = true;

            t.sFullType = t.sFullNakedType = t.sNestedType = t.sNakedType = t.sType;
        }

        return t;
    }

    //Log(lModWords);

    s_uint uiNumOpen;
    s_uint uiStartTemp;
    s_uint uiNumClosed;
    s_uint uiEndTemp;
    s_bool bTemplate;
    foreach (iter, lModWords)
    {
        if (*iter == "<")
        {
            ++uiNumOpen;
            if (uiStartTemp == 0)
                uiStartTemp = iter - lModWords.Begin();

            bTemplate = true;
        }
        else if (*iter == ">")
        {
            ++uiNumClosed;
            uiEndTemp = iter - lModWords.Begin();
        }
    }

    if (uiNumClosed > uiNumOpen)
        ThrowError("ill formed type, missing opening '<'");
    if (uiNumClosed < uiNumOpen)
        ThrowError("ill formed type, missing closing '>'");
    if (uiStartTemp > uiEndTemp)
        ThrowError("ill formed type, mismatched brackets");

    s_ctnr<s_str> lQualifiers;
    s_ctnr<s_str> lType;
    foreach (iter, lModWords)
    {
        if (bTemplate && s_uint((uint)(iter - lModWords.Begin())).IsInRange(uiStartTemp, uiEndTemp))
            continue;

        if (!lTypeQualifierList_.Find(*iter))
            lType.PushBack(*iter);
        else
            lQualifiers.PushBack(*iter);
    }

    //Log("type : " + lType);
    //Log("qual : " + lQualifiers);

    if (uiNumOpen != 0)
    {
        s_ctnr<s_ctnr<s_str>> lTemplateParams;
        s_ctnr<s_str> lParam;
        for (s_uint i = uiStartTemp+1; i < uiEndTemp; ++i)
        {
            if (lModWords[i] == ",")
            {
                if (lParam.IsEmpty())
                    ThrowError("ill formed type, empty template parameter");

                lTemplateParams.PushBack(lParam);
                lParam.Clear();
            }
            else
                lParam.PushBack(lModWords[i]);
        }

        if (uiEndTemp != uiStartTemp+1)
        {
            if (lParam.IsEmpty())
                ThrowError("ill formed type, empty template parameter");

            lTemplateParams.PushBack(lParam);
        }

        //Log("params : "+lTemplateParams);

        t.sTemplateParams = "<";

        s_ctnr<s_ctnr<s_str>>::iterator iterParam;
        foreach (iterParam, lTemplateParams)
        {
            TypeName tn = ReadType_(*iterParam);

            if (t.sTemplateParams == "<")
                t.sTemplateParams += tn.sFullType;
            else
                t.sTemplateParams += ", " + tn.sFullType;

            t.lTemplateParams.PushBack(tn);
        }

        t.sTemplateParams += ">";
    }

    if (lType.GetSize() == 1)
    {
        t.sType = lType[0];

        if (!lPODTypeList_.Find(t.sType) && !IsTemplateType_(t.sType))
        {
            if (!mState_.bNoTypeCheck_)
                mState_.pCurrentModule_->DecorateType(t, mState_.lContextStack_);
        }
        else
        {
            if (IsTemplateType_(t.sType))
                t.bIsTemplateParam = true;
            else if (!t.sTemplateParams.IsEmpty())
                ThrowError("POD types cannot have template parameters");

            t.sFullType = t.sNestedType = t.sFullNakedType = t.sNakedType = t.sType;
        }
    }
    else
    {
        foreach (iter, lType)
        {
            if (!lPODTypeList_.Find(*iter))
                ThrowError("ill formed type");

            if (t.sType.IsEmpty())
                t.sType = *iter;
            else
                t.sType += " " + *iter;
        }

        if (!t.sTemplateParams.IsEmpty())
            ThrowError("POD types cannot have template parameters");

        t.sFullType = t.sNestedType = t.sFullNakedType = t.sNakedType = t.sType;
    }

    t.lQualifiers = lQualifiers;

    if (!mState_.bNoTypeCheck_)
    {
        s_str sError = t.Qualify();
        if (!sError.IsEmpty())
            ThrowError(sError);
    }

    return t;
}

void Compiler::CheckOutputDirectory()
{
    if (!bOutputDirectoryChecked_)
    {
        if (!Directory::Exists(sOutputDirectory_))
            if (!Directory::Make(sOutputDirectory_))
                ThrowError("couldn't create output directory");

        bOutputDirectoryChecked_ = true;
    }
}
